"""A source for excel data"""

import json
import os
from pathlib import Path

import polars as pl

from ...diff.diff_model import Diff, DiffStats, DiffUpdate
from ..interface import SourceInterface
from ..registrated_source import register_source
from .excel_config import ExcelConfig, ExcelConfigArguments
from .excel_diff import ExcelDiff, ExcelDiffArguments
from .excel_sync import ExcelSync, ExcelSyncArguments


@register_source("Excel")
class ExcelSource(SourceInterface):
    """A source for excel data"""

    ConfigArgumentsClass = ExcelConfigArguments
    ConfigClass = ExcelConfig
    DiffArgumentsClass = ExcelDiffArguments
    DiffClass = ExcelDiff
    SyncClass = ExcelSync
    SyncArgumentsClass = ExcelSyncArguments

    connection_settings_arguments: ExcelConfigArguments = None
    connection_settings: ExcelConfig = None
    source_configuration_arguments: ExcelSyncArguments = None
    source_configuration: ExcelSync = None

    def _configure(self):
        pass

    def _connect(self):
        pass

    def _disconnect(self):
        pass

    def _query(self, params=None) -> pl.DataFrame:
        return pl.read_excel(
            self.connection_settings.path,
            schema_overrides=self.source_configuration.types,
            sheet_name=self.source_configuration.sheet_name,
        )

    def insert(self, data: pl.DataFrame):
        print("Inserting data into static source...")
        print(data)
        return len(data)

    def update(self, data: pl.DataFrame):
        print("Updating data in static source...")
        print(data)
        return len(data)

    def delete(self, data: pl.DataFrame):
        print("Deleting data from static source")
        print(data)
        return len(data)

    def store_diff(self, diff: Diff):
        existing_data = pl.DataFrame()

        if not os.path.exists(self.connection_settings.path):
            os.makedirs(os.path.dirname(self.connection_settings.path), exist_ok=True)
            Path(os.path.join(self.connection_settings.path)).touch()

        try:
            existing_data = pl.read_excel(self.connection_settings.path)
        # Broad catch is ok here since we just want to figure out if we need to create the file
        # pylint: disable=broad-exception-caught
        except Exception:
            # Do nothing, just replace the file
            pass

        new_data = pl.DataFrame(
            {
                "uuid": str(diff.uuid),
                "name": diff.name,
                "date": diff.date_as_string(),
                "version": diff.version,
                "updates": json.dumps(diff.updates, cls=DiffUpdate.JsonEncoder),
                "inserts": json.dumps(diff.inserts),
                "deletes": json.dumps(diff.deletes),
                "stats": json.dumps(diff.stats, cls=DiffStats.JsonEncoder),
            }
        )

        if existing_data.is_empty():
            existing_data = new_data
        else:
            existing_data = pl.concat([existing_data, new_data])

        existing_data.write_excel(self.connection_settings.path)



from typing import Literal

from ..interface.interface_diff import (
    SourceDiff,
    SourceDiffArguments,
    SourceDiffConfigArguments,
)


class XmlDiffConfigArguments(SourceDiffConfigArguments):
    pass


class XmlDiffArguments(SourceDiffArguments):
    type: Literal["Xml"]
    config: XmlDiffConfigArguments


class XmlDiff(SourceDiff):

    def __init__(self, diff_config: XmlDiffArguments) -> None:
        super().__init__(diff_config)



from polars import DataFrame
from .result import Result


class ComparisonResult:
    create: DataFrame
    update: DataFrame
    delete: DataFrame

    def __init__(self, create: DataFrame, update: DataFrame, delete: DataFrame):
        self.create = create
        self.update = update
        self.delete = delete

    def __str__(self):
        return f"Create: {self.create.height}, Update: {self.update.height}, Delete: {self.delete.height}"

    def __repr__(self):
        return str(self)

    def to_beetl_result(self):
        result = Result()
        result.inserts = self.create.height
        result.updates = self.update.height
        result.deletes = self.delete.height
        result.names = ("dry run",)
        return result



from typing import Annotated, Literal, Optional

from pydantic import Field

from ...typings import PolarTypeOverrides, PolarTypeOverridesParameters
from ..interface import SourceSync, SourceSyncArguments


class ExcelSyncArguments(SourceSyncArguments):
    types: Annotated[Optional[PolarTypeOverridesParameters], Field(default=None)]
    sheet_name: Annotated[Optional[str], Field(default=None)]
    type: Annotated[Literal["Excel"], Field(default="Excel")] = "Excel"


class ExcelSync(SourceSync):
    """The configuration class used for Excel file sources"""

    types: Optional[PolarTypeOverrides] = None
    sheet_name: Optional[str] = None

    def __init__(self, arguments: ExcelSyncArguments):
        super().__init__(arguments)

        self.sheet_name = arguments.sheet_name

        if arguments.types is not None:
            self.types = PolarTypeOverrides(arguments.types)



"""Contains the CsvSource class for handling CSV data sources."""

import json
import os
from pathlib import Path

import polars as pl

from ...diff import Diff, DiffStats, DiffUpdate
from ..interface import SourceInterface
from ..registrated_source import register_source
from .csv_config import CsvConfig, CsvConfigArguments
from .csv_diff import CsvDiff, CsvDiffArguments


@register_source("Csv")
class CsvSource(SourceInterface):
    """Source for interacting with CSV files."""

    ConfigArgumentsClass = CsvConfigArguments
    ConfigClass = CsvConfig
    DiffArgumentsClass = CsvDiffArguments
    DiffClass = CsvDiff

    diff_config_arguments: CsvDiffArguments = None
    diff_config: CsvDiff = None

    def _configure(self):
        pass

    def _connect(self):
        pass

    def _disconnect(self):
        pass

    def _query(self, params=None) -> pl.DataFrame:
        return pl.read_csv(
            self.connection_settings.path, encoding=self.connection_settings.encoding
        )

    def insert(self, data: pl.DataFrame):
        print("Inserting data into static source...")
        print(data)
        return len(data)

    def update(self, data: pl.DataFrame):
        print("Updating data in static source...")
        print(data)
        return len(data)

    def delete(self, data: pl.DataFrame):
        print("Deleting data from static source")
        print(data)
        return len(data)

    def store_diff(self, diff: Diff):
        if not self.diff_config:
            raise ValueError("Diff configuration is missing")

        existing_data = pl.DataFrame()

        if not os.path.exists(self.connection_settings.path):
            os.makedirs(os.path.dirname(self.connection_settings.path), exist_ok=True)
            Path(os.path.join(self.connection_settings.path)).touch()

        try:
            existing_data = pl.read_csv(
                self.connection_settings.path,
                encoding=self.connection_settings.encoding,
            )
        # Ok to catch broadly here since we are just figuring out if we should replace the file or not.
        # pylint: disable=broad-exception-caught
        except Exception:
            # Do nothing, just replace the file
            pass

        new_data = pl.DataFrame(
            {
                "uuid": str(diff.uuid),
                "name": diff.name,
                "date": diff.date_as_string(),
                "version": diff.version,
                "updates": json.dumps(diff.updates, cls=DiffUpdate.JsonEncoder),
                "inserts": json.dumps(diff.inserts),
                "deletes": json.dumps(diff.deletes),
                "stats": json.dumps(diff.stats, cls=DiffStats.JsonEncoder),
            }
        )

        if existing_data.is_empty():
            existing_data = new_data
        else:
            existing_data = pl.concat([existing_data, new_data])

        existing_data.write_csv(self.connection_settings.path)



"""Contains the diff configuration for the Postgresql source."""

from typing import Literal

from ..interface.interface_diff import (
    SourceDiff,
    SourceDiffArguments,
    SourceDiffConfigArguments,
)


class PostgresDiffConfigArguments(SourceDiffConfigArguments):
    """Represents the source specifig config for the diff configuration of the Postgresql source."""

    table: str


class PostgresDiffArguments(SourceDiffArguments):
    """Represents the arguments for the diff configuration for the Postgresql source."""

    type: Literal["Postgresql"]
    config: PostgresDiffConfigArguments


class PostgresDiff(SourceDiff):
    """Represents the parsed diff configuration for the Postgresql source."""

    table: str

    def __init__(self, diff_config: PostgresDiffArguments) -> None:
        super().__init__(diff_config)
        self.table = diff_config.config.table



from typing import Annotated, Literal

from pydantic import BaseModel, ConfigDict, Field


class SourceDiffConfigArguments(BaseModel):
    model_config = ConfigDict(extra="forbid")


class SourceDiffArguments(BaseModel):
    """Class representation of the diff configuration settings present in a beetl sync. Used to validate the diff configuration settings using pydantic.

    How to use:
    When deriving from this class you need to override the literal of the type field with the registered source name.

    You can then choose to override the default config field if your source requires additional configuration settings.
    """

    model_config = ConfigDict(extra="forbid")

    type: Literal["Interface"]
    name: Annotated[str, Field(min_length=1)]
    config: SourceDiffConfigArguments


class SourceDiff:
    def __init__(self, arguments: SourceDiffConfigArguments):
        pass



import json

from polars import DataFrame

from ...diff import Diff
from ..interface import SourceInterface
from ..registrated_source import register_source
from .static_config import StaticConfig, StaticConfigArguments
from .static_diff import StaticDiff, StaticDiffArguments


@register_source("Static")
class StaticSource(SourceInterface):
    ConfigArgumentsClass = StaticConfigArguments
    ConfigClass = StaticConfig
    DiffArgumentsClass = StaticDiffArguments
    DiffClass = StaticDiff

    """ A source for static data """

    def _configure(self):
        pass

    def _connect(self):
        pass

    def _disconnect(self):
        pass

    def _query(self, params=None) -> DataFrame:
        return self.connection_settings.data

    def insert(self, data: DataFrame):
        print("Inserting data into static source...")
        print(data)
        return len(data)

    def update(self, data: DataFrame):
        print("Updating data in static source...")
        print(data)
        return len(data)

    def delete(self, data: DataFrame):
        print("Deleting data from static source")
        print(data)
        return len(data)

    def store_diff(self, data: Diff):
        print("Storing diff in static source...")
        print(data.dump_json())



"""A source for fake data"""

from polars import DataFrame

from ...diff.diff_model import Diff
from ..interface import SourceInterface
from ..registrated_source import register_source
from .faker_config import FakerConfig, FakerConfigArguments
from .faker_diff import FakerDiff, FakerDiffArguments


@register_source("Faker")
class FakerSource(SourceInterface):
    """A source for fake data"""

    ConfigArgumentsClass = FakerConfigArguments
    ConfigClass = FakerConfig
    DiffArgumentsClass = FakerDiffArguments
    DiffClass = FakerDiff

    def _configure(self):
        pass

    def _connect(self):
        pass

    def _disconnect(self):
        pass

    def _query(self, params=None) -> DataFrame:
        return self.connection_settings.data

    def insert(self, data: DataFrame):
        print("Printing insert data to console...")
        print(data)
        return len(data)

    def update(self, data: DataFrame):
        print("Printing update data to console...")
        print(data)
        return len(data)

    def delete(self, data: DataFrame):
        print("Printing delete data to console...")
        print(data)
        return len(data)

    def store_diff(self, diff: Diff):
        print("Printing diff to console...")
        print(diff.dump_json())



import os
from typing import Literal

# re-export
from .csv import CsvConfig, CsvConfigArguments, CsvDiff, CsvDiffArguments, CsvSource
from .excel import (
    ExcelConfig,
    ExcelConfigArguments,
    ExcelDiff,
    ExcelDiffArguments,
    ExcelDiffConfigArguments,
    ExcelSource,
    ExcelSyncArguments,
)
from .faker import (
    FakerConfig,
    FakerConfigArguments,
    FakerDiff,
    FakerDiffArguments,
    FakerDiffConfigArguments,
    FakerSource,
)
from .interface import (
    SourceConfig,
    SourceConfigArguments,
    SourceConnectionArguments,
    SourceDiff,
    SourceDiffArguments,
    SourceDiffConfigArguments,
    SourceSync,
    SourceSyncArguments,
)
from .itop import ItopConfig, ItopConfigArguments, ItopSource, ItopSyncArguments
from .mongodb import (
    MongodbConfig,
    MongodbConfigArguments,
    MongodbDiff,
    MongodbDiffArguments,
    MongodbDiffConfigArguments,
    MongodbSource,
    MongodbSync,
    MongodbSyncArguments,
)
from .mysql import (
    MysqlConfig,
    MysqlConfigArguments,
    MysqlDiff,
    MysqlDiffArguments,
    MysqlDiffConfigArguments,
    MysqlSource,
    MysqlSyncArguments,
)
from .postgresql import (
    PostgresConfig,
    PostgresConfigArguments,
    PostgresDiff,
    PostgresDiffArguments,
    PostgresSource,
    PostgresSyncArguments,
)
from .registrated_source import Sources, register_source
from .request_threader import RequestThreader
from .rest import RestConfig, RestConfigArguments, RestSource, RestSyncArguments
from .sqlserver import (
    SqlserverConfig,
    SqlserverConfigArguments,
    SqlserverDiff,
    SqlserverDiffArguments,
    SqlserverDiffConfigArguments,
    SqlserverSource,
    SqlserverSync,
    SqlserverSyncArguments,
)
from .static import (
    StaticConfig,
    StaticConfigArguments,
    StaticDiff,
    StaticDiffArguments,
    StaticSource,
)
from .xml import (
    XmlConfig,
    XmlConfigArguments,
    XmlDiff,
    XmlDiffArguments,
    XmlSource,
    XmlSync,
    XmlSyncArguments,
)

cDir = os.path.dirname(__file__)
for file in os.listdir(cDir):
    if file.endswith(".py") and not file.startswith("__") and file != "interface.py":
        __import__(f"{__name__}.{file[:-3]}", fromlist=["*"])



import json
from dataclasses import dataclass
from typing import Any, Dict, List, Literal

import polars as pl
import yaml

from ..sources.interface import SourceConfig, SourceInterface, SourceSync
from ..transformers.interface import TransformerConfiguration
from ..typings import ComparisonColumn


class SourceSettings:
    """The configuration class used for datasources"""

    name: str
    source_type: str
    connection: SourceConfig
    config: SourceSync

    def __init__(
        self, name: str, source_type: str, connection: dict, config: dict
    ) -> None:
        self.name, self.source_type = name, source_type
        source_class = __import__(f"beetl.sources.{source_type.lower()}")

        self.connection = getattr(
            source_class, f"{source_type}SourceConnectionSettings"
        )(**connection)

        self.config = getattr(source_class, f"{source_type}SourceConfig")(**config)


@dataclass
class ChangeDetectionConfig:
    threshold: int = -1
    stopSync: bool = False
    webhook: str = None
    webhookHeaders: Dict[str, str] = None
    webhookPayload: Dict[str, str] = None


@dataclass
class SyncConfiguration:
    """The configuration for a single sync between two sources"""

    source: SourceInterface
    sourceConfig: SourceSync
    destination: SourceInterface
    destinationConfig: SourceSync
    comparisonColumns: list[ComparisonColumn]
    name: str = ""
    changeDetection: ChangeDetectionConfig = None

    sourceTransformers: TransformerConfiguration = None
    destinationTransformers: TransformerConfiguration = None

    insertionTransformers: list[TransformerConfiguration] = None
    deletionTransformers: list[TransformerConfiguration] = None

    diff_destination_instance: SourceInterface = None
    diff_transformers: list[TransformerConfiguration] = None

    def __post_init__(self) -> None:
        self.source.config = self.sourceConfig
        self.destination.config = self.destinationConfig


class BeetlConfig:
    """The configuration class for BeETL"""

    version: Literal["V1"]
    sources: Dict[str, SourceSettings]
    sync_list: List[SyncConfiguration]

    def __init__(self, config: dict) -> None:
        # Set version to keep track of changes
        # in an effort to keep backwards compatibility

        # Get the class corresponding to the configuration version
        module_root = str.join(".", self.__module__.split(".")[:-1])
        module = __import__(module_root, fromlist=[""])
        config_class = getattr(
            module, f'BeetlConfig{config.get("configVersion", "V1")}'
        )

        # Transform this class into the versioned configuration class
        self.__class__ = config_class
        self.__dict__ = config_class(config).__dict__
        self.version = config.get("configVersion", "V1")

    @classmethod
    def from_yaml(cls, yamlData: str) -> "BeetlConfig":
        """Import configuration from a YAML string

        Args:
            yaml (str): The YAML string containing the configuration

        Returns:
            BeetlConfig: Beetl Configuration
        """
        config = yaml.safe_load(yamlData)
        return cls(config)

    @classmethod
    def from_yaml_file(cls, path: str, encoding: str = "utf-8") -> "BeetlConfig":
        """Import configuration from YAML file

        Args:
            path (str): The path to the configuration file
            encoding (str, optional): The encoding of the file. Defaults to 'utf-8'.

        Raises:
            Exception: File not found

        Returns:
            BeetlConfig: Beetl Configuration
        """
        try:
            with open(path, "r", encoding=encoding) as file:
                config = yaml.safe_load(file)
                return cls(config)
        except FileNotFoundError as e:
            raise Exception(
                f"The configuration file was not found at the path specified ({path})."
            ) from e

    @classmethod
    def from_json(cls, jsonData: str) -> "BeetlConfig":
        """Import configuration from a JSON string

        Args:
            jsonData (str): The JSON string containing the configuration

        Returns:
            BeetlConfig: Beetl Configuration
        """
        config = json.loads(jsonData)
        return cls(config)

    @classmethod
    def from_json_file(cls, path: str, encoding: str = "utf-8") -> "BeetlConfig":
        """Import configuration from YAML file

        Args:
            path (str): The path to the configuration file
            encoding (str, optional): The encoding of the file. Defaults to 'utf-8'.

        Raises:
            Exception: File not found

        Returns:
            BeetlConfig: Beetl Configuration
        """
        try:
            with open(path, "r", encoding=encoding) as file:
                config = json.load(file)
                return cls(config)
        except FileNotFoundError as e:
            raise Exception(
                f"The configuration file was not found at the path specified ({path})."
            ) from e



from .mysql_config import MysqlConfig, MysqlConfigArguments
from .mysql_diff import MysqlDiff, MysqlDiffArguments, MysqlDiffConfigArguments
from .mysql_source import MysqlSource
from .mysql_sync import MysqlSync, MysqlSyncArguments



from typing import Literal

from ..interface.interface_diff import (
    SourceDiff,
    SourceDiffArguments,
    SourceDiffConfigArguments,
)


class MongodbDiffConfigArguments(SourceDiffConfigArguments):
    collection: str


class MongodbDiffArguments(SourceDiffArguments):
    type: Literal["Mongodb"]
    config: MongodbDiffConfigArguments


class MongodbDiff(SourceDiff):
    collection: str

    def __init__(self, diff_config: MongodbDiffArguments) -> None:
        super().__init__(diff_config)
        self.collection = diff_config.config.collection



from .comparison_column import ComparisonColumn
from .polar_types import (
    CASTABLE,
    PolarTypeLiterals,
    PolarTypeOverrides,
    PolarTypeOverridesParameters,
    get_polar_type_from_literal,
)



import json
import os
from pathlib import Path
from typing import Optional

import pandas as pd
import polars as pl

from ...diff import DiffStats, DiffUpdate
from ..interface import SourceInterface
from ..registrated_source import register_source
from .xml_config import XmlConfig, XmlConfigArguments
from .xml_diff import XmlDiff, XmlDiffArguments
from .xml_sync import XmlSync, XmlSyncArguments


@register_source("Xml")
class XmlSource(SourceInterface):
    """A source for reading and writing XML files"""

    ConfigArgumentsClass = XmlConfigArguments
    ConfigClass = XmlConfig
    SyncArgumentsClass = XmlSyncArguments
    SyncClass = XmlSync
    DiffArgumentsClass = XmlDiffArguments
    DiffClass = XmlDiff

    connection_settings: XmlConfig = None
    source_configuration: XmlSync = None
    mutation_data: pl.DataFrame = pl.DataFrame()
    diff_config: Optional[XmlDiff] = None
    diff_config_arguments: Optional[XmlDiffArguments] = None

    def _configure(self):
        pass

    def _connect(self):
        pass

    def _disconnect(self):
        """Flushes the mutated dataframe to the xml file"""
        self.flush_data_to_file(self.mutation_data)

    def _query(self, params=None) -> pl.DataFrame:
        """Reads the XML file and returns a polars DataFrame"""

        file_does_not_exist = not os.path.exists(self.connection_settings.path)
        if file_does_not_exist:
            return pl.DataFrame()

        file_is_empty = os.path.getsize(self.connection_settings.path) == 0
        if file_is_empty:
            return pl.DataFrame()

        data = pl.from_pandas(
            pd.read_xml(
                self.connection_settings.path,
                encoding=self.connection_settings.encoding,
                namespaces=self.source_configuration.namespaces,
                parser=self.source_configuration.parser,
                xpath=self.source_configuration.xpath,
                dtype=self.source_configuration.types,
                stylesheet=self.source_configuration.xsl,
            )
        )

        return data

    def insert(self, data: pl.DataFrame):
        """Xml files are not yet supported as a destination"""
        unique_columns = self.get_unique_columns()
        destination = self.load_data_for_mutation(data)
        updated = destination.update(
            data, on=unique_columns, how="full", include_nulls=True
        )
        self.save_data_after_mutation(updated)
        return len(data)

    def update(self, data: pl.DataFrame):
        """Xml files are not yet supported as a destination"""
        unique_columns = self.get_unique_columns()
        destination = self.load_data_for_mutation(data)
        updated = destination.update(
            data, on=unique_columns, how="left", include_nulls=True
        )
        self.save_data_after_mutation(updated)
        return len(data)

    def delete(self, data: pl.DataFrame):
        """Xml files are not yet supported as a destination"""
        unique_columns = self.get_unique_columns()
        destination = self.load_data_for_mutation(data)
        updated = destination.join(data, on=unique_columns, how="anti")
        self.save_data_after_mutation(updated)
        return len(data)

    def load_data_for_mutation(self, data_schema: pl.DataFrame) -> pl.DataFrame:
        if self.mutation_data.height > 0:
            return self.mutation_data

        mutation_data = self._query()
        self.mutation_data = self.append_missing_columns(data_schema, mutation_data)

        return self.mutation_data

    def save_data_after_mutation(self, data: pl.DataFrame) -> None:
        self.mutation_data = data

    def flush_data_to_file(self, data: pl.DataFrame) -> None:
        data_frame_is_missing_columns = data.width == 0
        if data_frame_is_missing_columns:
            return

        if not os.path.exists(self.connection_settings.path):
            with open(self.connection_settings.path, "w") as file:
                file.flush()

        data.to_pandas().to_xml(
            self.connection_settings.path,
            encoding=self.connection_settings.encoding,
            root_name=self.source_configuration.root_name,
            row_name=self.source_configuration.row_name,
            index=False,
            parser="etree",
        )

    def get_unique_columns(self):
        if not self.source_configuration.unique_columns:
            raise ValueError(
                "Unique columns must be defined in the configuration when used as a destination"
            )

        return self.source_configuration.unique_columns

    def append_missing_columns(
        self, source: pl.DataFrame, destination: pl.DataFrame
    ) -> pl.DataFrame:
        columns_to_append = [
            column for column in source.columns if column not in destination.columns
        ]
        for column in columns_to_append:
            destination = destination.with_columns(
                pl.Series(
                    column, [None] * destination.height, dtype=source[column].dtype
                )
            )

        return destination

    def store_diff(self, diff):
        existing_data = pl.DataFrame()

        if not os.path.exists(self.connection_settings.path):
            os.makedirs(os.path.dirname(self.connection_settings.path), exist_ok=True)
            Path(os.path.join(self.connection_settings.path)).touch()

        try:
            # The index is automatically added when reading XML files with pandas, to be able to merge the dataframes we need to drop it.
            # It will be automatically recreated when writing the file again.
            existing_data = pl.from_pandas(
                pd.read_xml(self.connection_settings.path)
            ).drop(["index"])
        # OK to catch all exceptions here as we're just deciding whether we need to create the file
        # pylint: disable=broad-exception-caught
        except Exception:
            # Do nothing, just replace the file
            pass

        new_data = pl.DataFrame(
            {
                "uuid": str(diff.uuid),
                "name": diff.name,
                "date": diff.date_as_string(),
                "version": diff.version,
                "updates": json.dumps(diff.updates, cls=DiffUpdate.JsonEncoder),
                "inserts": json.dumps(diff.inserts),
                "deletes": json.dumps(diff.deletes),
                "stats": json.dumps(diff.stats, cls=DiffStats.JsonEncoder),
            }
        )

        if existing_data.is_empty():
            existing_data = new_data
        else:
            existing_data = pl.concat([existing_data, new_data])

        existing_data.to_pandas().to_xml(self.connection_settings.path)



from typing import Literal

from ..interface.interface_diff import (
    SourceDiff,
    SourceDiffArguments,
    SourceDiffConfigArguments,
)


class MysqlDiffConfigArguments(SourceDiffConfigArguments):
    table: str


class MysqlDiffArguments(SourceDiffArguments):
    type: Literal["Mysql"]
    config: MysqlDiffConfigArguments


class MysqlDiff(SourceDiff):
    table: str

    def __init__(self, diff_config: MysqlDiffArguments) -> None:
        super().__init__(diff_config)
        self.table = diff_config.config.table



from time import perf_counter
from typing import List, Union

from tabulate import tabulate

from .compare.compare import Difftool
from .comparison_result import ComparisonResult
from .config import BeetlConfig
from .diff import DiffCalculator
from .result import Result, SyncResult
from .transformers import run_transformers

BENCHMARK = []


class Beetl:
    """The main class for BeETL. This class is responsible for orchestrating the ETL process."""

    config: Union[BeetlConfig, None] = None
    """Holds the BeETL Configuration"""

    def __init__(self, config: BeetlConfig):
        self.config = config

    @classmethod
    def from_yaml(cls, path: str, encoding: str = "utf-8") -> "Beetl":
        """Creates a Beetl instance from a YAML-formatted configuration file

        Args:
            path (str): The path to the file
            encoding (str, optional): The charset for the file. Defaults to 'utf-8'.

        Returns:
            Beetl: An instance of Beetl
        """
        return cls(BeetlConfig.from_yaml_file(path, encoding))

    @classmethod
    def from_json(cls, path: str, encoding: str = "utf-8") -> "Beetl":
        """Creates a Beetl instance from a JSON-formatted configuration file

        Args:
            path (str): The path to the file
            encoding (str, optional): The charset for the file. Defaults to 'utf-8'.

        Returns:
            Beetl: _description_
        """
        return cls(BeetlConfig.from_json_file(path, encoding))

    def benchmark(self, text: str) -> None:
        """Inserts a benchmark into the log"""
        BENCHMARK.append({"text": text, "perf": perf_counter()})

        if len(BENCHMARK) > 1:
            print(
                BENCHMARK[-1]["text"]
                + ": "
                + str(round(BENCHMARK[-1]["perf"] - BENCHMARK[-2]["perf"], 5))
            )

    def sync(
        self, dry_run: bool = False, generate_update_diff: bool = False
    ) -> Union[Result, List[ComparisonResult]]:
        """Executes the ETL process.

        Args:
            dry_run (bool, optional): If set to true, the function will not execute any database operations and will return the dataframes that would have been used in the operations. Defaults to False.
            generate_update_diff (bool, optional): If set to true, the function will return the differences between the source and destination update dataframes. Defaults to False.

        Returns:
            ComparisonResult: If the argument dry_run was passed as true, returns a list of ComparisionResult objects that contain the create, update and delete dataframes that would have been used by the destination if the dry_run argument was false.
            Result: A Result object containing the amount of inserts, updates and deletes for each sync
            list[tuple[list[DataFrame], set[str]]]: If generate_update_diff is set to true, returns a list of tuples containing the differences between the source and destination update dataframes and a set with the keys that have differences across the whole dataframe.

        The following steps will be performed:

        1. Load source and destination data

        2. Format source data to be compatible with destination through field and source transformers

        3. Compare source and destination data with compare_datasets

        4. Execute the respective insert, update and delete queries

        """
        dry_run_results = []
        generate_update_diff_results = []
        self.benchmark("Starting sync and retrieving source data")
        allAmounts = []
        for i, sync in enumerate(self.config.sync_list, 1):
            start = perf_counter()
            if sync.name != "":
                print(f"Starting sync: {sync.name}")
            else:
                print(f"Starting sync {i}")
            sync.source.connect()
            source_data = sync.source.query(sync.sourceConfig)
            sync.source.disconnect()
            self.benchmark("Finished data retrieval from source")

            sync.destination.connect()
            destination_data = sync.destination.query(sync.destinationConfig)
            self.benchmark("Finished data retrieval from destination")

            self.benchmark("Starting source data transformation")
            transformedSource = run_transformers(source_data, sync.sourceTransformers)
            self.benchmark(
                "Finished source data transformation, starting destination transformation"
            )
            transformedDestination = run_transformers(
                destination_data, sync.destinationTransformers
            )

            self.benchmark("Finished data transformation before comparison")

            self.benchmark("Starting comparison")
            unique_columns = tuple(
                column.name for column in sync.comparisonColumns if column.unique
            )
            comparison_columns = tuple(
                column.name
                for column in sync.comparisonColumns
                if column.name not in unique_columns
            )

            if len(unique_columns) == 0:
                raise ValueError(
                    "You need to specify at least one unique column in the sync.comparisonColumns field"
                )

            diff_calculator = DiffCalculator(
                sync.name,
                transformedSource,
                transformedDestination,
                sync.comparisonColumns,
            )
            create, update, delete = diff_calculator.get_create_update_delete_for_sync()
            self.benchmark("Successfully extracted operations from dataset")

            load_and_compare = perf_counter() - start
            print(f"Load and compare took {load_and_compare} seconds")

            amount = {}

            print(
                f"Insert: {len(create)}, Update: {len(update)}, Delete: {len(delete)}"
            )

            if generate_update_diff:
                generate_update_diff_results.append(
                    Difftool.diff_update(
                        update,
                        transformedDestination,
                        comparison_columns,
                        unique_columns,
                    )
                )
                continue

            if dry_run:
                dry_run_results.append(
                    ComparisonResult(
                        run_transformers(create, sync.insertionTransformers),
                        run_transformers(update, sync.insertionTransformers),
                        run_transformers(delete, sync.deletionTransformers),
                    )
                )
                sync.destination.disconnect()
                continue

            if sync.diff_destination_instance is not None:
                diff = diff_calculator.create_diff(sync.diff_transformers)
                sync.diff_destination_instance.connect()
                sync.diff_destination_instance.store_diff(diff)

            self.benchmark("Starting database operations")
            self.benchmark("Starting deletes")
            amount["deletes"] = 0
            if len(delete):
                amount["deletes"] = sync.destination.delete(
                    run_transformers(delete, sync.deletionTransformers)
                )

            self.benchmark("Finished deletes, starting inserts")
            amount["inserts"] = 0
            if len(create):
                amount["inserts"] = sync.destination.insert(
                    run_transformers(create, sync.insertionTransformers)
                )

            self.benchmark("Finished inserts, starting updates")

            amount["updates"] = 0
            if len(update):
                amount["updates"] = sync.destination.update(
                    run_transformers(update, sync.insertionTransformers)
                )

            self.benchmark("Finished updates, sync finished")

            print("Inserted: " + str(amount["inserts"]))
            print("Updated: " + str(amount["updates"]))
            print("Deleted: " + str(amount["deletes"]))

            allAmounts.append(
                [sync.name, *[amount["inserts"], amount["updates"], amount["deletes"]]]
            )

            sync.destination.disconnect()
            if sync.diff_destination_instance:
                sync.diff_destination_instance.disconnect()

        if generate_update_diff:
            return generate_update_diff_results

        if dry_run:
            return dry_run_results

        print(
            "\r\n\r\n"
            + tabulate(allAmounts, headers=["Sync", "Inserts", "Updates", "Deletes"])
        )

        return SyncResult(allAmounts)



from .excel_config import ExcelConfig, ExcelConfigArguments
from .excel_diff import ExcelDiff, ExcelDiffArguments, ExcelDiffConfigArguments
from .excel_source import ExcelSource
from .excel_sync import ExcelSyncArguments



"""Classes related to comparison column definition."""

from dataclasses import dataclass
from typing import Any

import polars as pl


@dataclass
class ComparisonColumn:
    """
    A class representing a column used for comparison, with attributes for its name, data type, and uniqueness.

    Attributes:
      name (str): The name of the column.
      type (pl.DataType): The data type of the column, determined dynamically from a string representation.
      unique (bool): Indicates whether the column values are unique. Defaults to False.
    """

    name: str
    type: pl.DataType
    unique: bool = False

    def __init__(self, name: str, type: Any, unique: bool = False) -> None:
        self.name, self.unique = name, unique
        self.type = getattr(pl, type)



"""Utilities related to calculating the diff between two datasets."""

from typing import Literal, Optional

import polars as pl
from polars import DataFrame

from ..constants import BEETL_COMPARISON_HASH_COLUMN_IDENTIFIER, RESERVED_IDENTIFIERS
from ..transformers import TransformerConfiguration, run_transformers
from ..typings import CASTABLE, ComparisonColumn
from .diff_model import Diff, DiffRow, DiffUpdate


class DiffCalculator:
    """
    DiffCalculator is a utility class designed to compute the differences
      between two datasets represented as DataFrames.
    It identifies changes based on unique and comparison
      columns specified during initialization.
    """

    sync_name: str
    unique_columns: tuple[str, ...]
    comparison_columns: tuple[str, ...]
    source: DataFrame
    destination: DataFrame

    updates_old: Optional[DataFrame] = None
    updates_new: Optional[DataFrame] = None
    updates_diff_mask: Optional[DataFrame] = None

    inserts: Optional[DataFrame] = None
    deletes: Optional[DataFrame] = None

    def __init__(
        self,
        sync_name: str,
        source: DataFrame,
        destination: DataFrame,
        columns: list[ComparisonColumn],
    ):
        self.sync_name = sync_name
        self.source = self._cast_columns_to_specified_types(
            self._initialize_columns_if_empty(source, columns), columns
        )
        self.destination = self._cast_columns_to_specified_types(
            self._initialize_columns_if_empty(destination, columns), columns
        )
        self.unique_columns = [col.name for col in columns if col.unique]
        self.comparison_columns = [
            col.name for col in columns if col.name not in self.unique_columns
        ]

        (
            self.inserts,
            self.updates_old,
            self.updates_new,
            self.updates_diff_mask,
            self.deletes,
        ) = self._calculate_diff()

    def _calculate_diff(self) -> tuple[DataFrame, DataFrame, DataFrame]:
        columns_contain_list_column = any(
            [
                True
                for name in self.comparison_columns
                if isinstance(self.source.get_column(name).dtype, pl.List)
            ]
        )
        if columns_contain_list_column:
            raise ValueError(
                """Beetl does not support comparing list columns.
                Please remove them from the sync.comparisonColumns field.
                """
            )

        inserts = self._calculate_inserts(
            self.source, self.destination, self.unique_columns
        )
        deletes = self._calculate_deletes(
            self.source, self.destination, self.unique_columns
        )

        if self.diff_cannot_contain_any_updates():
            return (inserts, DataFrame(), DataFrame(), DataFrame(), deletes)

        updates_old, updates_new, updated_diff_mask = self._calculate_updates(
            self.source, self.destination, self.unique_columns, self.comparison_columns
        )

        return (inserts, updates_old, updates_new, updated_diff_mask, deletes)

    def diff_cannot_contain_any_updates(self):
        """
        Nothing can change if there are no comparison columns.
        The only thing that can change is the unique columns.
        That means that rows will be inserted or deleted, but not updated.

        returns:
            bool: True if there are no comparison columns.
        """
        return len(self.comparison_columns) == 0

    @staticmethod
    def _calculate_inserts(
        source: DataFrame,
        destination: DataFrame,
        unique_columns: tuple[str, ...],
    ) -> DataFrame:
        return source.join(destination, on=unique_columns, how="anti")

    @staticmethod
    def _calculate_deletes(
        source: DataFrame,
        destination: DataFrame,
        unique_columns: tuple[str, ...],
    ) -> DataFrame:
        return destination.join(source, on=unique_columns, how="anti")

    @staticmethod
    def _calculate_updates(
        source: DataFrame,
        destination: DataFrame,
        unique_columns: tuple[str, ...],
        comparison_columns: tuple[str, ...],
    ) -> tuple[DataFrame, DataFrame]:
        source_filtered = source.select(*unique_columns, *comparison_columns)
        destination_filtered = destination.select(*unique_columns, *comparison_columns)

        destination_renamed_as_old = destination_filtered.rename(
            {
                col: f"{col}_old"
                for col in destination_filtered.columns
                if col not in unique_columns
            }
        )

        merged = source_filtered.join(
            destination_renamed_as_old, on=unique_columns, how="outer"
        )

        update_diff_mask_expression = [
            (merged[col] != merged[f"{col}_old"]).alias(f"diff_{col}")
            for col in comparison_columns
        ]

        merged = merged.with_columns(update_diff_mask_expression)
        updated_rows = merged.filter(pl.any_horizontal(update_diff_mask_expression))
        diff_mask = updated_rows.select(
            *unique_columns, *[f"diff_{col}" for col in comparison_columns]
        )
        updated_rows_identifiers = updated_rows.select(*unique_columns)

        old = destination.join(updated_rows_identifiers, on=unique_columns, how="inner")
        new = source.join(updated_rows_identifiers, on=unique_columns, how="inner")

        return (old, new, diff_mask)

    @staticmethod
    def _initialize_columns_if_empty(
        dataframe: DataFrame, columns: tuple[ComparisonColumn, ...]
    ) -> DataFrame:
        if len(dataframe) == 0 and dataframe.width == 0:
            for col in columns:
                dataframe = dataframe.with_columns(pl.Series(col.name, dtype=col.type))
        return dataframe

    @staticmethod
    def _cast_columns_to_specified_types(
        dataframe: DataFrame, columns: tuple[ComparisonColumn, ...]
    ) -> DataFrame:
        for col in columns:
            if col.name in dataframe.columns and col.type in CASTABLE:
                dataframe = dataframe.with_columns(pl.col(col.name).cast(col.type))
        return dataframe

    def create_diff(
        self, transformers: Optional[list[TransformerConfiguration]] = None
    ) -> Diff:
        """
        Creates a diff object that represents the differences between two datasets.

        This method calculates the differences by comparing the deletes, inserts,
        and updates between the old and new datasets. It applies optional
        transformations to the datasets before performing the comparison.

        Args:
            transformers (Optional[list[TransformerConfiguration]]):
                A list of transformer configurations to apply to the datasets
                before computing the diff. Defaults to None.

        Returns:
            Diff: An object containing the differences, including updates, inserts,
            and deletes, between the old and new datasets.
        """
        diff_deletes = tuple(
            map(
                DiffRow,
                run_transformers(self.deletes, transformers).to_dicts(),
            )
        )
        diff_inserts = tuple(
            map(
                DiffRow,
                run_transformers(self.inserts, transformers).to_dicts(),
            )
        )

        add_hash_expression = (
            pl.struct(self.unique_columns)
            .hash()
            .alias(BEETL_COMPARISON_HASH_COLUMN_IDENTIFIER)
        )
        old_with_hash = self.updates_old.with_columns(add_hash_expression)
        new_with_hash = self.updates_new.with_columns(add_hash_expression)

        old_transformed = run_transformers(old_with_hash, transformers)
        new_transformed = run_transformers(new_with_hash, transformers)

        columns = old_transformed.columns

        old_renamed = old_transformed.rename(
            self._create_rename_mapping("old", columns)
        )
        new_renamed = new_transformed.rename(
            self._create_rename_mapping("new", columns)
        )

        updates = old_renamed.join(
            new_renamed, on=BEETL_COMPARISON_HASH_COLUMN_IDENTIFIER, how="inner"
        )

        updates = updates.with_columns(
            self._create_struct_column_expression("new", columns)
        )
        updates = updates.with_columns(
            self._create_struct_column_expression("old", columns)
        )

        updates = updates.select(["old", "new"])

        diff_updates = []
        for update_row in updates.to_dicts():
            old = DiffRow(update_row["old"])
            new = DiffRow(update_row["new"])
            diff_updates.append(DiffUpdate(old, new))

        return Diff(self.sync_name, diff_updates, diff_inserts, diff_deletes)

    @staticmethod
    def _create_rename_mapping(
        postfix: Literal["old", "new"], column_names: tuple[str, ...]
    ):
        return {
            col: f"{col}_{postfix}"
            for col in column_names
            if col not in RESERVED_IDENTIFIERS
        }

    @staticmethod
    def _create_struct_column_expression(
        postfix: Literal["old", "new"], columns: tuple[str, ...]
    ):
        return pl.struct(
            [
                pl.col(f"{col}_{postfix}").alias(col)
                for col in columns
                if col not in RESERVED_IDENTIFIERS
            ]
        ).alias(postfix)

    def get_create_update_delete_for_sync(
        self,
    ) -> tuple[DataFrame, DataFrame, DataFrame]:
        """
        Retrieves the DataFrames representing the inserts, updates, and deletes
        required for synchronization.

        Returns:
            tuple[DataFrame, DataFrame, DataFrame]: A tuple containing three DataFrames:
                - The first DataFrame represents the rows to be inserted.
                - The second DataFrame represents the rows to be updated.
                - The third DataFrame represents the rows to be deleted.
        """
        return (self.inserts, self.updates_new, self.deletes)



from typing import Literal

from ..interface.interface_diff import (
    SourceDiff,
    SourceDiffArguments,
    SourceDiffConfigArguments,
)


class SqlserverDiffConfigArguments(SourceDiffConfigArguments):
    table: str


class SqlserverDiffArguments(SourceDiffArguments):
    type: Literal["Sqlserver"]
    config: SqlserverDiffConfigArguments


class SqlserverDiff(SourceDiff):
    table: str

    def __init__(self, diff_config: SqlserverDiffArguments) -> None:
        self.table = diff_config.config.table



"""
Definition for the SqlServer source
"""

import json
from typing import Optional
from uuid import uuid4

import pandas as pd
import polars as pl
import pyodbc
import sqlalchemy as sqla

from ...diff import Diff, DiffStats, DiffUpdate
from ..interface import SourceInterface
from ..registrated_source import register_source
from .sqlserver_config import SqlserverConfig, SqlserverConfigArguments
from .sqlserver_diff import SqlserverDiff, SqlserverDiffArguments
from .sqlserver_sync import SqlserverSync, SqlserverSyncArguments


@register_source("Sqlserver")
class SqlserverSource(SourceInterface):
    """
    Class for interacting with SqlServer databases.
    """

    ConfigArgumentsClass = SqlserverConfigArguments
    ConfigClass = SqlserverConfig
    SyncArgumentsClass = SqlserverSyncArguments
    SyncClass = SqlserverSync
    DiffArgumentsClass = SqlserverDiffArguments
    DiffClass = SqlserverDiff

    diff_config_arguments: Optional[SqlserverDiffArguments] = None
    diff_config: Optional[SqlserverDiff] = None
    connection: Optional[pyodbc.Connection] = None
    engine: Optional[sqla.Engine] = None

    def _configure(self):
        pass

    def _connect(self):
        try:
            self.engine = sqla.create_engine(self.connection_settings.connection_string)
        except ModuleNotFoundError:
            try:
                self.engine = sqla.create_engine(
                    self.connection_settings.connection_string.replace(
                        "mssql://", "mssql+pyodbc://"
                    )
                )
            except ModuleNotFoundError:
                self.engine = sqla.create_engine(
                    self.connection_settings.connection_string.replace(
                        "mssql://", "mssql+pymssql://"
                    )
                )
        self.connection = self.engine.connect()

    def _disconnect(self):
        self.connection.commit()
        self.connection.close()

    def _query(
        self, params=None, customQuery: str = None, returnData: bool = True
    ) -> pl.DataFrame:
        query = customQuery

        if query is None:
            if self.source_configuration.query is not None:
                query = self.source_configuration.query

            if query is None:
                if self.source_configuration.table is None:
                    raise Exception("No query or table specified")

                query = f"SELECT * FROM [{self.source_configuration.table}]"

        if returnData:
            pdd = pd.read_sql_query(sql=query, con=self.connection)
            return pl.from_pandas(pdd)

        if isinstance(query, str):
            query = sqla.text(query)
        self.connection.execute(query)

    def _insert(
        self,
        data: pl.DataFrame,
        table: str = None,
        tempDB: bool = False,
    ):
        # Check for binary tables
        data_columns = data.get_columns()
        for column in data_columns:
            if column.dtype == pl.Binary:
                data = data.with_columns(
                    pl.col(column.name).map_elements(
                        lambda x: bytes(x), return_dtype=pl.Binary
                    )
                )

        # Replace empty strings with None
        if self.source_configuration.replace_empty_strings:
            for column_name in data.columns:
                if data[column_name].dtype in (pl.Utf8, pl.String):
                    data = data.with_columns(
                        pl.when(pl.col(column_name).eq(""))
                        .then(None)
                        .otherwise(pl.col(column_name))
                        .alias(column_name)
                    )

        if table is None:
            table = self.source_configuration.table

        pand_df = data.to_pandas()
        try:
            pand_df.to_sql(
                table,
                if_exists="append" if not tempDB else "replace",
                index=False,
                con=self.connection,
            )
        except ModuleNotFoundError:
            data.write_database(
                table,
                connection=self.connection,
                if_exists="append" if not tempDB else "replace",
            )

        return len(data)

    def insert(self, data: pl.DataFrame):
        if len(data) == 0:
            return 0

        return self._insert(data)

    # Help function to handle COLLATE
    def _collate_clause(
        self, column_name: str, schema: dict, left_alias: str, right_alias: str
    ):
        dtype = schema.get(column_name)

        if dtype in (pl.Utf8, pl.String):
            return (
                f"TRY_CONVERT(NVARCHAR, {left_alias}.[{column_name}]) COLLATE DATABASE_DEFAULT = "
                f"TRY_CONVERT(NVARCHAR, {right_alias}.[{column_name}]) COLLATE DATABASE_DEFAULT"
            )
        else:
            return f"{left_alias}.[{column_name}] = " f"{right_alias}.[{column_name}]"

    def update(self, data: pl.DataFrame):
        self._validate_unique_columns()
        if len(data) == 0:
            return 0

        tempDB = self._get_temp_table_name(self.source_configuration.table)

        # Insert data into temporary table
        self._insert(data, table=tempDB, tempDB=True)

        columns_to_update = [
            column_name
            for column_name in data.columns
            if column_name not in self.source_configuration.unique_columns
            and column_name not in self.source_configuration.skip_columns
        ]
        field_spec = ", ".join(
            (
                f"TDEST.[{column_name}] = TTEMP.[{column_name}]"
                for column_name in columns_to_update
            )
        )

        schema = data.schema
        on_clause = " AND ".join(
            self._collate_clause(column_name, schema, "TDEST", "TTEMP")
            for column_name in self.source_configuration.unique_columns
        )

        query = f"""
            UPDATE TDEST
            SET {field_spec}
            FROM {self.source_configuration.table} AS TDEST
            INNER JOIN {tempDB} AS TTEMP ON {on_clause}
        """
        self._query(customQuery=query, returnData=False)
        try:
            self._query(customQuery="DROP TABLE " + tempDB, returnData=False)
        except Exception:
            pass

        return len(data)

    def delete(self, data: pl.DataFrame):
        self._validate_unique_columns()
        if len(data) == 0:
            return 0

        tempDB = self._get_temp_table_name(self.source_configuration.table)
        self._insert(data, table=tempDB, tempDB=True)

        schema = data.schema
        where_clause = " AND ".join(
            (
                self._collate_clause(column_name, schema, "TTEMP", "TDEST")
                for column_name in self.source_configuration.unique_columns
            )
        )

        if self.source_configuration.soft_delete:
            if (
                self.source_configuration.deleted_field is None
                or self.source_configuration.deleted_value is None
            ):
                raise Exception(
                    "Deleted field and value must be specified when using soft delete"
                )

            query = f"""
                UPDATE TDEST
                SET {self.source_configuration.deleted_field} = {self.source_configuration.deleted_value}
                FROM {self.source_configuration.table} AS TDEST
                WHERE EXISTS (
                    SELECT 1/0
                    FROM {tempDB} AS TTEMP
                    WHERE {where_clause}
                ) 
            """
        else:
            query = f"""
                DELETE TDEST
                FROM {self.source_configuration.table} AS TDEST
                WHERE EXISTS (
                    SELECT 1/0
                    FROM {tempDB} AS TTEMP
                    WHERE {where_clause}
                ) 
            """

        self._query(customQuery=query, returnData=False)

        try:
            self._query(customQuery="DROP TABLE " + tempDB, returnData=False)
        except Exception:
            pass

        return len(data)

    def _validate_unique_columns(self):
        if not self.source_configuration.unique_columns:
            raise ValueError(
                "Unique columns must be specified for SQLServer source when used as destination"
            )

    def _get_temp_table_name(self, table_name: str):
        return f"##{table_name}_{str(uuid4()).replace('-', '')}_temp".lower()

    def store_diff(self, diff: Diff):
        if not self.diff_config:
            raise ValueError("Diff configuration is missing")
        metadata = sqla.MetaData()
        table = sqla.Table(
            self.diff_config.table,
            metadata,
            sqla.Column("uuid", sqla.Uuid, primary_key=True),
            sqla.Column("name", sqla.String),
            sqla.Column("date", sqla.DateTime),
            sqla.Column("version", sqla.String),
            sqla.Column("updates", sqla.String),
            sqla.Column("inserts", sqla.String),
            sqla.Column("deletes", sqla.String),
            sqla.Column("stats", sqla.String),
        )
        metadata.create_all(self.connection)

        insert_statement = sqla.insert(table).values(
            name=diff.name,
            date=diff.date,
            uuid=diff.uuid,
            version=diff.version,
            updates=json.dumps(diff.updates, cls=DiffUpdate.JsonEncoder),
            inserts=json.dumps(diff.inserts),
            deletes=json.dumps(diff.deletes),
            stats=json.dumps(diff.stats, cls=DiffStats.JsonEncoder),
        )
        self.connection.execute(insert_statement)



from typing import Annotated, Literal

from pydantic import Field

from ..interface.interface_diff import SourceDiff, SourceDiffArguments


class StaticDiffArguments(SourceDiffArguments):
    type: Literal["Static"]


class StaticDiff(SourceDiff):
    pass



"""Resources related to running transformers"""

import polars as pl

from .interface import TransformerConfiguration


def run_transformers(
    source: pl.DataFrame,
    transformers: list[TransformerConfiguration],
) -> pl.DataFrame:
    """
    Applies a list of transformers to a source DataFrame.
    """
    transformed = source.clone()

    if transformers is not None and len(transformers) > 0:
        for transformer in transformers:
            if transformer.include_sync:
                transformed = transformer.transform(transformed)
                continue

            transformed = transformer.transform(transformed)

    return transformed



from typing import Literal

from ..interface.interface_diff import (
    SourceDiff,
    SourceDiffArguments,
    SourceDiffConfigArguments,
)


class ExcelDiffConfigArguments(SourceDiffConfigArguments):
    pass


class ExcelDiffArguments(SourceDiffArguments):
    type: Literal["Excel"]
    config: ExcelDiffConfigArguments


class ExcelDiff(SourceDiff):

    def __init__(self, diff_config: ExcelDiffArguments) -> None:
        super().__init__(diff_config)



from .mongodb_config import MongodbConfig, MongodbConfigArguments
from .mongodb_diff import MongodbDiff, MongodbDiffArguments, MongodbDiffConfigArguments
from .mongodb_source import MongodbSource
from .mongodb_sync import MongodbSync, MongodbSyncArguments



"""A source for MySQL data"""

import json
from typing import Any

import polars as pl
import sqlalchemy as sqla

from ...diff import DiffStats, DiffUpdate
from ..interface import SourceInterface
from ..registrated_source import register_source
from .mysql_config import MysqlConfig, MysqlConfigArguments
from .mysql_diff import MysqlDiff, MysqlDiffArguments
from .mysql_sync import MysqlSync, MysqlSyncArguments


@register_source("Mysql")
class MysqlSource(SourceInterface):
    """A source for MySQL data"""

    ConfigArgumentsClass = MysqlConfigArguments
    ConfigClass = MysqlConfig
    SyncArgumentsClass = MysqlSyncArguments
    SyncClass = MysqlSync
    DiffArgumentsClass = MysqlDiffArguments
    DiffClass = MysqlDiff

    diff_config: MysqlDiff = None
    diff_config_arguments: MysqlDiffArguments = None

    def _configure(self):
        pass

    def _connect(self):
        pass

    def _disconnect(self):
        pass

    def _query(
        self, params=None, customQuery: str = None, returnData: bool = True
    ) -> pl.DataFrame:
        query = customQuery

        if query is None:
            if self.source_configuration.query is not None:
                query = self.source_configuration.query

            if query is None:
                if self.source_configuration.table is None:
                    raise ValueError("No query or table specified")

                query = f"SELECT * FROM {self.source_configuration.table}"

        if returnData:
            try:
                df = pl.read_database_uri(
                    query=query,
                    uri=self.connection_settings.connection_string,
                )
            except ModuleNotFoundError:
                df = pl.read_database_uri(
                    query=query,
                    uri=self.connection_settings.connection_string.replace(
                        "mysql://", "mysql+pymysql://"
                    ),
                )

            return df
        try:
            with sqla.create_engine(
                self.connection_settings.connection_string
            ).connect() as con:
                con.execute(query)
                con.commit()

        except ModuleNotFoundError:
            with sqla.create_engine(
                self.connection_settings.connection_string.replace(
                    "mysql://", "mysql+pymysql://"
                )
            ).connect() as con:
                con.execute(sqla.text(query))
                con.commit()

    def _insert(
        self, data: pl.DataFrame, table: str = None, connection_string: str = None
    ):
        if table is None:
            table = self.source_configuration.table
        if connection_string is None:
            connection_string = self.connection_settings.connection_string

        try:
            data.write_database(
                table,
                self.connection_settings.connection_string,
                if_table_exists="append",
            )
        except ModuleNotFoundError:
            data.write_database(
                table,
                self.connection_settings.connection_string.replace(
                    "mysql://", "mysql+pymysql://"
                ),
                if_table_exists="append",
            )

        return len(data)

    def insert(self, data: pl.DataFrame):
        self._validate_unique_columns()
        return self._insert(data)

    def update(self, data: pl.DataFrame):
        self._validate_unique_columns()

        tempDB = (self.source_configuration.table + "_udTemp").lower()
        try:
            with sqla.create_engine(
                self.connection_settings.connection_string
            ).connect() as con:
                con.execute(sqla.text(f"DROP TABLE IF EXISTS {tempDB}"))
                con.execute(
                    sqla.text(
                        f"CREATE TEMPORARY TABLE {tempDB} "
                        f"LIKE {self.source_configuration.table}"
                    )
                )
                con.commit()

        except ModuleNotFoundError:
            with sqla.create_engine(
                self.connection_settings.connection_string.replace(
                    "mysql://", "mysql+pymysql://"
                )
            ).connect() as con:
                con.execute(sqla.text(f"DROP TABLE IF EXISTS {tempDB}"))
                con.execute(
                    sqla.text(
                        f"CREATE TEMPORARY TABLE {tempDB} "
                        f"LIKE {self.source_configuration.table}"
                    )
                )
                con.commit()

        # Insert data into temporary table
        self._insert(data, table=tempDB)

        fields_to_update = [
            field.name
            for field in data.get_columns()
            if not field.name in self.source_configuration.skip_columns
            or not field.name in self.source_configuration.unique_columns
        ]

        field_spec = ", ".join((f"`{fieldName}`" for fieldName in fields_to_update))

        query = f"""
            REPLACE INTO {self.source_configuration.table} ({field_spec})
            SELECT {field_spec} FROM {tempDB}
        """

        self._query(customQuery=query, returnData=False)

        return len(data)

    def delete(self, data: pl.DataFrame):
        batch_size = 500
        batches = [data]

        if len(data) > batch_size:
            batches = []

            for i in range(0, len(data), batch_size):
                batches.append(data[i : i + batch_size])

        for batch in batches:
            id_clause = " AND ".join(
                (
                    f"`{fieldName}` IN ({','.join([self._quote_if_needed(x) for x in batch[fieldName].to_list()])})"
                    for fieldName in self.source_configuration.unique_columns
                )
            )

            query = f"""
                DELETE FROM {self.source_configuration.table}
                WHERE {id_clause}
            """

            self._query(customQuery=query, returnData=False)

        return len(data)

    def _validate_unique_columns(self):
        if not self.source_configuration.unique_columns:
            raise ValueError(
                "MySQL source requires the unique_columns to be set if used as a destination"
            )

    def _quote_if_needed(self, identifier: Any) -> str:
        if isinstance(identifier, str):
            return f"'{identifier}'"
        return str(identifier)

    def store_diff(self, diff):
        if not self.diff_config:
            raise ValueError("Diff configuration is missing")
        metadata = sqla.MetaData()
        table = sqla.Table(
            self.diff_config.table,
            metadata,
            sqla.Column("uuid", sqla.Uuid, primary_key=True),
            sqla.Column("name", sqla.String(length=255)),
            sqla.Column("date", sqla.DateTime),
            sqla.Column("version", sqla.String(length=64)),
            sqla.Column("updates", sqla.TEXT),
            sqla.Column("inserts", sqla.TEXT),
            sqla.Column("deletes", sqla.TEXT),
            sqla.Column("stats", sqla.TEXT),
        )

        insert_statement = sqla.insert(table).values(
            name=diff.name,
            date=diff.date,
            uuid=diff.uuid,
            version=diff.version,
            updates=json.dumps(diff.updates, cls=DiffUpdate.JsonEncoder),
            inserts=json.dumps(diff.inserts),
            deletes=json.dumps(diff.deletes),
            stats=json.dumps(diff.stats, cls=DiffStats.JsonEncoder),
        )
        with sqla.create_engine(
            self.connection_settings.connection_string.replace(
                "mysql://", "mysql+pymysql://"
            )
        ).connect() as con:
            metadata.create_all(con)
            con.execute(insert_statement)
            con.commit()



from typing import Literal

from ..interface.interface_diff import (
    SourceDiff,
    SourceDiffArguments,
    SourceDiffConfigArguments,
)


class CsvDiffConfigArguments(SourceDiffConfigArguments):
    pass


class CsvDiffArguments(SourceDiffArguments):
    type: Literal["Csv"]
    config: CsvDiffConfigArguments


class CsvDiff(SourceDiff):

    def __init__(self, diff_config: CsvDiffArguments) -> None:
        super().__init__(diff_config)



# re-export
from .postgresql_config import PostgresConfig, PostgresConfigArguments
from .postgresql_diff import PostgresDiff, PostgresDiffArguments
from .postgresql_source import PostgresSource
from .postgresql_sync import PostgresSync, PostgresSyncArguments



import json
from datetime import datetime
from json import JSONEncoder
from typing import Any, Literal
from uuid import uuid4

DiffRow = dict[str, Any]


class DiffUpdate:
    old: DiffRow
    new: DiffRow

    def __init__(self, old: DiffRow, new: DiffRow):
        self.old = old
        self.new = new

    class JsonEncoder(JSONEncoder):
        def default(self, o: Any):
            if isinstance(o, DiffUpdate):
                return {"old": o.old, "new": o.new}

            return super().default(o)

    def dump_json(self):
        return json.dumps(self, cls=JSONEncoder)

    def to_dict(self) -> dict[str, DiffRow]:
        return {
            "old": self.old,
            "new": self.new,
        }


class DiffStats:
    updates: int
    inserts: int
    deletes: int

    def __init__(self, updates: int, inserts: int, deletes: int):
        self.updates, self.inserts, self.deletes = (
            len(updates),
            len(inserts),
            len(deletes),
        )

    class JsonEncoder(JSONEncoder):
        def default(self, o: Any):
            if isinstance(o, DiffStats):
                return {
                    "updates": o.updates,
                    "inserts": o.inserts,
                    "deletes": o.deletes,
                }
            return super().default(o)

    def dump_json(self):
        return json.dumps(self, cls=JSONEncoder)

    def to_dict(self) -> dict[str, int]:
        return {
            "updates": self.updates,
            "inserts": self.inserts,
            "deletes": self.deletes,
        }


class Diff:
    name: str
    date: datetime
    uuid: str
    version: Literal["1.0.0"] = "1.0.0"
    updates: tuple[DiffUpdate, ...]
    inserts: tuple[DiffRow, ...]
    deletes: tuple[DiffRow, ...]
    stats: DiffStats

    def __init__(
        self,
        name: str,
        updates: tuple[DiffUpdate, ...],
        inserts: tuple[DiffRow, ...],
        deletes: tuple[DiffRow, ...],
    ):
        self.name = name
        self.date = datetime.now()
        self.uuid = uuid4()
        self.updates = updates
        self.inserts = inserts
        self.deletes = deletes
        self.stats = DiffStats(updates, inserts, deletes)

    def dump_json(self):
        return json.dumps(self, cls=DiffJsonEncoder)

    def date_as_string(self) -> str:
        return self.date.isoformat()


class DiffJsonEncoder(JSONEncoder):
    def default(self, o: Any):
        if isinstance(o, Diff):
            return {
                "name": o.name,
                "date": o.date.isoformat(),
                "uuid": str(o.uuid),
                "version": o.version,
                "updates": [
                    DiffUpdate.JsonEncoder().default(update) for update in o.updates
                ],
                "inserts": o.inserts,
                "deletes": o.deletes,
                "stats": DiffStats.JsonEncoder().default(o.stats),
            }
        return super().default(o)



# re-export
from .sqlserver_config import SqlserverConfig, SqlserverConfigArguments
from .sqlserver_diff import (
    SqlserverDiff,
    SqlserverDiffArguments,
    SqlserverDiffConfigArguments,
)
from .sqlserver_source import SqlserverSource
from .sqlserver_sync import SqlserverSync, SqlserverSyncArguments
